@@ -24,6 +24,7 @@ import dis
 import os
 import platform
 import signal
+import struct
 import subprocess
 import sys
 import threading
@@ -81,7 +82,6 @@ class adaptive:
     def add(self, value: float) -> None:
         if self.current_index >= self.max_samples:
             # Decimate
-            # print("DECIMATION " + str(self.sample_array))
             new_array = [0.0] * self.max_samples
             for i in range(0, self.max_samples // 3):
                 arr = [self.sample_array[i * 3 + j] for j in range(0, 3)]
@@ -89,7 +89,6 @@ class adaptive:
                 new_array[i] = arr[1]  # Median
             self.current_index = self.max_samples // 3
             self.sample_array = new_array
-            # print("POST DECIMATION = " + str(self.sample_array))
             # Update average length
             self.max_average *= 3
         self.sample_array[self.current_index] = value
@@ -183,10 +182,10 @@ def parse_args() -> Tuple[argparse.Namespace, List[str]]:
         prog="scalene",
         description=usage,
         formatter_class=argparse.RawTextHelpFormatter,
+        allow_abbrev=False,
     )
     parser.add_argument("prog", type=str, help="program to be profiled")
     parser.add_argument(
-        "-o",
         "--outfile",
         type=str,
         default=None,
@@ -222,6 +221,10 @@ def parse_args() -> Tuple[argparse.Namespace, List[str]]:
         default=False,
         help="only profile CPU time (default: profile CPU, memory, and copying)",
     )
+    # the PID of the profiling process (for internal use only)
+    parser.add_argument(
+        "--pid", type=int, default=int(os.getpid()), help=argparse.SUPPRESS
+    )
     # Parse out all Scalene arguments and jam the remaining ones into argv.
     # https://stackoverflow.com/questions/35733262/is-there-any-way-to-instruct-argparse-python-2-7-to-remove-found-arguments-fro
     args, left = parser.parse_known_args()
@@ -233,13 +236,19 @@ arguments, left = parse_args()
 # Load shared objects unless the user specifies "--cpu-only" at the command-line.
 # (x86-64 only for now.)
 
-if not arguments.cpuonly and platform.machine() != "x86_64":
+if not arguments.cpuonly and (
+    platform.machine() != "x86_64" or struct.calcsize("P") * 8 != 64
+):
     arguments.cpuonly = True
     print(
-        "scalene warning: currently only x86-64 platforms are supported for memory and copy profiling."
+        "scalene warning: currently only 64-bit x86-64 platforms are supported for memory and copy profiling."
     )
 
-if not arguments.cpuonly and platform.machine() == "x86_64":
+if (
+    not arguments.cpuonly
+    and platform.machine() == "x86_64"
+    and struct.calcsize("P") * 8 == 64
+):
     # Load the shared object on Linux.
     if sys.platform == "linux":
         if ("LD_PRELOAD" not in os.environ) and ("PYTHONMALLOC" not in os.environ):
@@ -257,13 +266,15 @@ if not arguments.cpuonly and platform.machine() == "x86_64":
         if ("DYLD_INSERT_LIBRARIES" not in os.environ) and (
             "PYTHONMALLOC" not in os.environ
         ):
-            os.environ["DYLD_INSERT_LIBRARIES"] = os.path.join(
+            env = os.environ
+            # qenv = {}
+            env["DYLD_INSERT_LIBRARIES"] = os.path.join(
                 os.path.dirname(__file__), "libscalene.dylib"
             )
-            os.environ["PYTHONMALLOC"] = "malloc"
+            env["PYTHONMALLOC"] = "malloc"
             args = sys.argv[1:]
             args = [os.path.basename(sys.executable), "-m", "scalene"] + args
-            result = subprocess.run(args)
+            result = subprocess.run(args, env=env, close_fds=True, shell=False)
             sys.exit(result.returncode)
 
 Filename = NewType("Filename", str)
@@ -485,7 +496,7 @@ class Scalene:
     __memcpy_signal = signal.SIGPROF
 
     # Whether we are in a signal handler or not (to make things properly re-entrant).
-    __in_signal_handler = 0
+    __in_signal_handler = threading.RLock()
 
     # We cache the previous signal handlers so we can play nice with
     # apps that might already have handlers for these signals.
@@ -584,7 +595,7 @@ start the timer interrupts."""
         Scalene.__original_memcpy_signal_handler = signal.signal(
             Scalene.__memcpy_signal, Scalene.memcpy_event_signal_handler
         )
-        # Turn on the CPU profiling timer to run every signal_interval seconds.
+        # Turn on the CPU profiling timer to run every mean_signal_interval seconds.
         signal.setitimer(
             Scalene.__cpu_timer_signal,
             Scalene.__mean_signal_interval,
@@ -654,9 +665,10 @@ process."""
             # Hijack lock.
             threading.Lock = Scalene.ReplacementLock  # type: ignore
             # Hijack system and subprocess calls.
-            os.system = Scalene.new_os_system  # type: ignore
-            os.popen = Scalene.new_os_popen  # type: ignore
-            subprocess.Popen = Scalene.new_subprocess_Popen  # type: ignore
+            # os.system = Scalene.new_os_system  # type: ignore
+            # os.popen = Scalene.new_os_popen  # type: ignore
+            # subprocess.Popen = Scalene.new_subprocess_Popen  # type: ignore
+            pass
         # Build up signal filenames (adding PID to each).
         Scalene.__malloc_signal_filename = Filename(
             Scalene.__malloc_signal_filename + str(os.getpid())
@@ -675,15 +687,21 @@ process."""
 
     @staticmethod
     def cpu_signal_handler(
+        signum: Union[Callable[[Signals, FrameType], None], int, Handlers, None],
+        this_frame: FrameType,
+    ) -> None:
+        if Scalene.__in_signal_handler.acquire(blocking=False):
+            Scalene.cpu_signal_handler_helper(signum, this_frame)
+            Scalene.__in_signal_handler.release()
+
+    @staticmethod
+    def cpu_signal_handler_helper(
         _signum: Union[Callable[[Signals, FrameType], None], int, Handlers, None],
         this_frame: FrameType,
     ) -> None:
         """Handle interrupts for CPU profiling."""
         # Record how long it has been since we received a timer
         # before.  See the logic below.
-        if Scalene.__in_signal_handler > 0:
-            return
-        Scalene.__in_signal_handler += 1
         now = Scalene.gettime()
         # If it's time to print some profiling info, do so.
         if now >= Scalene.__next_output_time:
@@ -777,7 +795,6 @@ process."""
         #        Scalene.__last_signal_interval,
         #    )
         Scalene.__last_signal_time = Scalene.gettime()
-        Scalene.__in_signal_handler -= 1
 
     # Returns final frame (up to a line in a file we are profiling), the thread identifier, and the original frame.
     @staticmethod
@@ -826,6 +843,7 @@ process."""
                     frame = cast(FrameType, frame.f_back)
                     if frame:
                         fname = frame.f_code.co_filename
+                        continue
             if frame:
                 new_frames.append((frame, tident, orig_frame))
         return new_frames
@@ -836,7 +854,9 @@ process."""
         this_frame: FrameType,
     ) -> None:
         """Handle malloc events."""
-        Scalene.allocation_handler(signum, this_frame)
+        if Scalene.__in_signal_handler.acquire(blocking=False):
+            Scalene.allocation_signal_handler(signum, this_frame)
+            Scalene.__in_signal_handler.release()
 
     @staticmethod
     def free_signal_handler(
@@ -844,22 +864,19 @@ process."""
         this_frame: FrameType,
     ) -> None:
         """Handle free events."""
-        Scalene.allocation_handler(signum, this_frame)
+        if Scalene.__in_signal_handler.acquire(blocking=False):
+            Scalene.allocation_signal_handler(signum, this_frame)
+            Scalene.__in_signal_handler.release()
 
     @staticmethod
-    def allocation_handler(
+    def allocation_signal_handler(
         signum: Union[Callable[[Signals, FrameType], None], int, Handlers, None],
         this_frame: FrameType,
     ) -> None:
-        if Scalene.__in_signal_handler > 0:
-            return
-        Scalene.__in_signal_handler += 1
-
         """Handle interrupts for memory profiling (mallocs and frees)."""
         new_frames = Scalene.compute_frames_to_record(this_frame)
 
         if not new_frames:
-            Scalene.__in_signal_handler -= 1
             return
 
         # Process the input array.
@@ -896,7 +913,7 @@ process."""
         # and update the global __memory_footprint_samples.
         before = Scalene.__current_footprint
         for item in arr:
-            alloc_time, action, count, python_fraction = item
+            _alloc_time, action, count, python_fraction = item
             # print(alloc_time, action, count, python_fraction)
             count /= 1024 * 1024
             is_malloc = action == "M"
@@ -913,7 +930,7 @@ process."""
         # This is a pain, since we don't know to whom to attribute memory,
         # so we may overcount.
 
-        for (frame, tident, orig_frame) in new_frames:
+        for (frame, _tident, _orig_frame) in new_frames:
             fname = Filename(frame.f_code.co_filename)
             line_no = LineNumber(frame.f_lineno)
             bytei = ByteCodeIndex(frame.f_lasti)
@@ -924,7 +941,7 @@ process."""
             allocs = 0.0
             # Go through the array again and add each updated current footprint.
             for item in arr:
-                alloc_time, action, count, python_fraction = item
+                _alloc_time, action, count, python_fraction = item
                 count /= 1024 * 1024
                 is_malloc = action == "M"
                 if is_malloc:
@@ -950,7 +967,6 @@ process."""
                 Scalene.__memory_free_samples[fname][line_no][bytei] += before - after
                 Scalene.__memory_free_count[fname][line_no][bytei] += 1
                 Scalene.__total_memory_free_samples += before - after
-        Scalene.__in_signal_handler -= 1
 
     @staticmethod
     def memcpy_event_signal_handler(
@@ -958,13 +974,11 @@ process."""
         frame: FrameType,
     ) -> None:
         """Handles memcpy events."""
-        if Scalene.__in_signal_handler > 0:
-            return
-        Scalene.__in_signal_handler += 1
-        new_frames = Scalene.compute_frames_to_record(frame)
-        if not new_frames:
-            Scalene.__in_signal_handler -= 1
-            return
+        if Scalene.__in_signal_handler.acquire(False):
+            new_frames = Scalene.compute_frames_to_record(frame)
+            if not new_frames:
+                Scalene.__in_signal_handler.release()
+                return
 
         # Process the input array.
         arr: List[Tuple[int, int]] = []
@@ -983,8 +997,8 @@ process."""
         arr.sort()
 
         for item in arr:
-            memcpy_time, count = item
-            for (the_frame, tident, orig_frame) in new_frames:
+            _memcpy_time, count = item
+            for (the_frame, _tident, _orig_frame) in new_frames:
                 fname = Filename(the_frame.f_code.co_filename)
                 line_no = LineNumber(the_frame.f_lineno)
                 bytei = ByteCodeIndex(the_frame.f_lasti)
@@ -992,7 +1006,7 @@ process."""
                 Scalene.__bytei_map[fname][line_no].add(bytei)
                 Scalene.__memcpy_samples[fname][line_no] += count
 
-        Scalene.__in_signal_handler -= 1
+        Scalene.__in_signal_handler.release()
 
     @staticmethod
     @lru_cache(None)
@@ -1306,7 +1320,7 @@ process."""
         else:
             if not Scalene.__output_file:
                 # No output file specified: write to stdout.
-                console.save_text("/dev/stdout", styles=True)
+                sys.stdout.write(console.export_text(styles=True))
             else:
                 # Don't output styles to text file.
                 console.save_text(Scalene.__output_file, styles=False, clear=False)
@@ -1326,7 +1340,8 @@ process."""
         Scalene.disable_signals()
 
     @contextlib.contextmanager
-    def scalene_profiler(_):  # type: ignore  ### TODO add memory stuff
+    @staticmethod
+    def scalene_profiler(_):  # type: ignore
         """A profiler function, work in progress."""
         # In principle, this would let people use Scalene as follows:
         # with scalene_profiler:
@@ -1375,7 +1390,7 @@ process."""
                         exec(code, the_globals, the_locals)
                     except BaseException:  # as be
                         # Intercept sys.exit.
-                        # print(traceback.format_exc())  # FIXME
+                        # print(traceback.format_exc()) # for debugging only
                         pass
                     profiler.stop()
                     # If we've collected any samples, dump them.
@@ -1393,7 +1408,7 @@ process."""
                     print(message)
                     print(traceback.format_exc())
         except (FileNotFoundError, IOError):
-            print("Scalene: could not find input file.")
+            print("Scalene: could not find input file " + args.prog)
 
 
 if __name__ == "__main__":
